#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>

#if defined(WIN32)
#  pragma warning(disable:4996)
#  include "glew.h"
#  include <GL/gl.h>
#elif defined(__APPLE__) || defined(MACOSX)
#  include <OpenGL/gl.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/gl.h>
#endif

#include "Obj.h"

/*
** ǥեȥ󥹥ȥ饯
*/
Obj::Obj()
{
  init();
}

/*
** ԡ󥹥ȥ饯
*/
Obj::Obj(const Obj &o)
{
  copy(o);
}

/*
** ǥȥ饯
*/
Obj::~Obj()
{
  if (vert) delete[] vert;
  if (norm) delete[] norm;
  if (fnorm) delete[] fnorm;
  if (face) delete[] face;
}

/*
** 黻
*/
Obj &Obj::operator=(const Obj &o)
{
  if (this != &o) {
    this->~Obj();
    copy(o);
  }
  return *this;
}

/*
** ֥Ȥν
*/
void Obj::init()
{
  nv = nf = 0;
  vert = norm = fnorm = 0;
  face = 0;
}

/*
** ֥ȤΥԡ
*/
void Obj::copy(const Obj &o)
{
  nv = o.nv;
  nf = o.nf;

  try {
    if (nv > 0) {
      vert = new Vec[nv];
      norm = new Vec[nv];

      memcpy(vert, o.vert, sizeof(Vec) * nv);
      memcpy(norm, o.norm, sizeof(Vec) * nv);
    }
    else {
      vert = norm = 0;
    }
    if (nf > 0) {
      fnorm = new Vec[nf];
      face = new Idx[nf];

      memcpy(fnorm, o.fnorm, sizeof(Vec) * nf);
      memcpy(face, o.face, sizeof(Idx) * nf);
    }
    else {
      fnorm = 0;
      face = 0;
    }
  }
  catch (std::bad_alloc e) {
    std::cerr << "꤬­ޤ"<< std::endl;
    this->~Obj();
    init();
  }
}

/*
** եɤ߹
*/
bool Obj::load(const char *name)
{
  /* γ */
  this->~Obj();

  /* եɤ߹ */
  std::ifstream file(name, std::ios::binary);
  if (!file) {
    std::cerr << name << " ޤ" << std::endl;
    init();
    return false;
  }

  /* ǡοĴ٤ */
  char buf[1024];
  int v, f;
  v = f = 0;
  while (file.getline(buf, sizeof buf)) {
    if (buf[0] == 'v' && buf[1] == ' ') {
      ++v;
    }
    else if (buf[0] == 'f' && buf[1] == ' ') {
      ++f;
    }
  }

  nv = v;
  nf = f;

  /* γ */
  try {
    vert = new Vec[v];
    norm = new Vec[v];
    fnorm = new Vec[f];
    face = new Idx[f];
  }
  catch (std::bad_alloc e) {
    std::cerr << "꤬­ޤ" << std::endl;
    this->~Obj();
    init();
    return false;
  }

  /* եδᤷ */
  file.clear();
  file.seekg(0L, std::ios::beg);

  /* ǡɤ߹ */
  v = f = 0;
  while (file.getline(buf, sizeof buf)) {
    if (buf[0] == 'v' && buf[1] == ' ') {
      sscanf(buf, "%*s %f %f %f", vert[v], vert[v] + 1, vert[v] + 2);
      ++v;
    }
    else if (buf[0] == 'f' && buf[1] == ' ') {
      if (sscanf(buf + 2, "%d/%*d/%*d %d/%*d/%*d %d/%*d/%*d", face[f], face[f] + 1, face[f] + 2) != 3) {
        if (sscanf(buf + 2, "%d//%*d %d//%*d %d//%*d", face[f], face[f] + 1, face[f] + 2) != 3) {
          sscanf(buf + 2, "%d %d %d", face[f], face[f] + 1, face[f] + 2);
        }
      }
      --face[f][0];
      --face[f][1];
      --face[f][2];
      ++f;
    }
  }

  /* ˡ٥ȥλ */
  int i;
  for (i = 0; i < f; ++i) {
    float dx1 = vert[face[i][1]][0] - vert[face[i][0]][0];
    float dy1 = vert[face[i][1]][1] - vert[face[i][0]][1];
    float dz1 = vert[face[i][1]][2] - vert[face[i][0]][2];
    float dx2 = vert[face[i][2]][0] - vert[face[i][0]][0];
    float dy2 = vert[face[i][2]][1] - vert[face[i][0]][1];
    float dz2 = vert[face[i][2]][2] - vert[face[i][0]][2];

    fnorm[i][0] = dy1 * dz2 - dz1 * dy2;
    fnorm[i][1] = dz1 * dx2 - dx1 * dz2;
    fnorm[i][2] = dx1 * dy2 - dy1 * dx2;
  }

  /* ĺβˡ٥ȥλ */
  for (i = 0; i < v; ++i) {
    norm[i][0] = norm[i][1] = norm[i][2] = 0.0;
  }
  
  for (i = 0; i < f; ++i) {
    norm[face[i][0]][0] += fnorm[i][0];
    norm[face[i][0]][1] += fnorm[i][1];
    norm[face[i][0]][2] += fnorm[i][2];

    norm[face[i][1]][0] += fnorm[i][0];
    norm[face[i][1]][1] += fnorm[i][1];
    norm[face[i][1]][2] += fnorm[i][2];

    norm[face[i][2]][0] += fnorm[i][0];
    norm[face[i][2]][1] += fnorm[i][1];
    norm[face[i][2]][2] += fnorm[i][2];
  }

  /* ĺβˡ٥ȥ */
  for (i = 0; i < v; ++i) {
    float a = sqrt(norm[i][0] * norm[i][0]
                 + norm[i][1] * norm[i][1]
                 + norm[i][2] * norm[i][2]);

    if (a != 0.0) {
      norm[i][0] /= a;
      norm[i][1] /= a;
      norm[i][2] /= a;
    }
  }

  return true;
}

/*
** ޷ɽ
*/
void Obj::draw(GLint posLocation, GLint vecLocation, GLsizei points) const
{
  /* ѷοݥꥴ */
  for (int i = 0; i < nf; ++i) {
    int i0 = face[i][0], i1 = face[i][1], i2 = face[i][2];

    /* uniform ѿ pos ˻ѷĺ p0 Ȥ飲ĺ˸٥ȥ e1, e2 Ǽ */
    GLfloat pos[] = {
      vert[i0][0], vert[i0][1], vert[i0][2],
      vert[i1][0] - vert[i0][0], vert[i1][1] - vert[i0][1], vert[i1][2] - vert[i0][2],
      vert[i2][0] - vert[i0][0], vert[i2][1] - vert[i0][1], vert[i2][2] - vert[i0][2],
    };
    glUniform3fv(posLocation, 3, pos);

    /* uniform ѿ vec ˻ѷΣĺˤˡ٥ȥǼ */
    GLfloat vec[] = {
      norm[i0][0], norm[i0][1], norm[i0][2],
      norm[i1][0], norm[i1][1], norm[i1][2],
      norm[i2][0], norm[i2][1], norm[i2][2],
    };
    glUniform3fv(vecLocation, 3, vec);

    /* ɽΰ褤äѤ˶ݥꥴ */
    glDrawArrays(GL_TRIANGLE_FAN, 0, points);
  }
}
