#include <cstdio>
#include <cstdlib>
#include <cmath>
#if defined(WIN32)
//#  pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#  pragma comment(lib, "glew32.lib")
#  include "glew.h"
#  include "glut.h"
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

/*
** ǡ
*/
#include "Obj.h"
static const char objectData[] = "bunny_s.obj";
static Obj *obj;

/*
** 
*/
#include "shadersource.h"
static const char vertSource[] = "simple.vert";
static const char fragSource[] = "simple.frag";
static GLuint shader;

/*
** 
*/
static const float cameraPosition[] = { 0.0f, 0.0f, 200.0f };
static const float cameraTarget[] = { 0.0f, 0.0f, 0.0f };
static const float cameraUp[] = { 0.0f, 1.0f, 0.0f };
static const float cameraFov = 60.0f;
static const float cameraNear = 1.0f;
static const float cameraFar = 1000.0f;

/*
** ɥ
*/
static int width, height;

/*
** θ
*/
static void lookat(const float *e, const float *t, const float *u, float *m)
{
  float tx, ty, tz, l;

  /* z  = e - t */
  tx = e[0] - t[0];
  ty = e[1] - t[1];
  tz = e[2] - t[2];
  l = sqrt(tx * tx + ty * ty + tz * tz);
  if (l > 0.0f) {
    m[6] = tx / l;
    m[7] = ty / l;
    m[8] = tz / l;
  }

  /* x  = u x z  */
  tx = u[1] * m[8] - u[2] * m[7];
  ty = u[2] * m[6] - u[0] * m[8];
  tz = u[0] * m[7] - u[1] * m[6];
	l = sqrt(tx * tx + ty * ty + tz * tz);
  if (l > 0.0f) {
    m[0] = tx / l;
    m[1] = ty / l;
    m[2] = tz / l;
  }

  /* y  = z  x x  */
  m[3] = m[7] * m[2] - m[8] * m[1];
  m[4] = m[8] * m[0] - m[6] * m[2];
  m[5] = m[6] * m[1] - m[7] * m[0];
}

/*
** ɽ
*/
static void display(void)
{
  /* ϻμ */
  int beginTime = glutGet(GLUT_ELAPSED_TIME);
  
  /* 賫 */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);

  /* ץͭˤ */
  glUseProgram(shader);

  /* ̤ȤεΥ 1 ȤȤ̤礭 */
  GLfloat sy = tan(cameraFov * 0.5f * 3.1415926f / 180.0f);
  GLfloat sx = sy * (float)width / (float)height;

  /* ̤̾Υ */
  GLfloat org[3];
  org[2] = cameraNear;
  org[0] = org[2] * sx;
  org[1] = org[2] * sy;

  /* ̤θ̾̾Ѱ */
  GLfloat dir[3];
  dir[2] = cameraFar - cameraNear;
  dir[0] = dir[2] * sx;
  dir[1] = dir[2] * sy;

  /* ž */
  GLfloat rot[9];
  lookat(cameraPosition, cameraTarget, cameraUp, rot);

  /* Сƥå uniform ѿξ */
  GLint camLocation = glGetUniformLocation(shader, "cam");
  GLint orgLocation = glGetUniformLocation(shader, "org");
  GLint dirLocation = glGetUniformLocation(shader, "dir");
  GLint rotLocation = glGetUniformLocation(shader, "rot");

  /* Сƥå uniform ѿͤꤹ */
  glUniform3fv(camLocation, 1, cameraPosition);
  glUniform3fv(orgLocation, 1, org);
  glUniform3fv(dirLocation, 1, dir);
  glUniformMatrix3fv(rotLocation, 1, GL_FALSE, rot);

  /* ɽΰ褤äѤݥꥴĺɸ */
  static const GLfloat scr[][3] = {
    { -1.0f, -1.0f, -1.0f },
    {  1.0f, -1.0f, -1.0f },
    {  1.0f,  1.0f, -1.0f },
    { -1.0f,  1.0f, -1.0f },
  };

  /* ĺХåե֥Ȥ */
  GLuint bufferObject;
  glGenBuffers(1, &bufferObject);

  /* ĺХåե֥Ȥ˥ǡ */
  glBindBuffer(GL_ARRAY_BUFFER, bufferObject);
  glBufferData(GL_ARRAY_BUFFER, sizeof scr, scr, GL_STATIC_DRAW);

  /* attribute ѿξ */
  GLint scrLocation = glGetAttribLocation(shader, "scr");

  /* attribute ѿ˥Хåե֥Ȥξꤹ */
  glVertexAttribPointer(scrLocation, sizeof scr[0] / sizeof scr[0][0], GL_FLOAT, GL_FALSE, 0, 0);

  /* attribute ѿͭˤ */
  glEnableVertexAttribArray(scrLocation);

  /* ե饰ȥ uniform ѿξ */
  GLint posLocation = glGetUniformLocation(shader, "pos");
  GLint vecLocation = glGetUniformLocation(shader, "vec");

  /* ݥꥴ */
  obj->draw(posLocation, vecLocation, sizeof scr / sizeof scr[0]);

  /* attribute ѿ̵ˤ */
  glDisableVertexAttribArray(scrLocation);

  /* ĺХåե֥Ȥ */
  glDeleteBuffers(1, &bufferObject);

  /* ץ̵ˤ */
  glUseProgram(0);

  /* 轪λ */
  glDisable(GL_DEPTH_TEST);
  glFinish();

  /* в֤ɽ */
  fprintf(stderr, "Elapsed time = %f\n", (double)(glutGet(GLUT_ELAPSED_TIME) - beginTime) * 0.001);
}

/*
** ɥ
*/
static void resize(int w, int h)
{
  /* ɥȹ⤵¸ */
  width = w;
  height = h;
  
  /* ɥΤӥ塼ݡȤˤ */
  glViewport(0, 0, w, h);
}

/*
** ܡ
*/
static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
    /* q  Q  ESC 򥿥פ齪λ */
    case '\033':
    case 'q':
    case 'Q':
      exit(0);
    case 'g':
    case 'G':
      /* g  G 򥿥פ */
      glutPostRedisplay();
    default:
      break;
  }
}

/*
** ץɤ߹
*/
static GLuint loadShader(const char *vertSource, const char *fragSource)
{
  /* ֥Ȥκ */
  GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
  GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
  
  /* Υץɤ߹ */
  if (readShaderSource(vertShader, vertSource)) return 0;
  if (readShaderSource(fragShader, fragSource)) return 0;
  
  /* ץΥѥ롿󥯷̤ѿ */
  GLint status;
  
  /* СƥåΥץΥѥ */
  glCompileShader(vertShader);
  glGetShaderiv(vertShader, GL_COMPILE_STATUS, &status);
  printShaderInfoLog(vertShader);
  if (status == GL_FALSE) {
    fprintf(stderr, "Compile error in %s.\n", vertSource);
    return 0;
  }
  
  /* ե饰ȥΥץΥѥ */
  glCompileShader(fragShader);
  glGetShaderiv(fragShader, GL_COMPILE_STATUS, &status);
  printShaderInfoLog(fragShader);
  if (status == GL_FALSE) {
    fprintf(stderr, "Compile error in %s.\n", fragSource);
    return 0;
  }
  
  /* ץ४֥Ȥκ */
  GLuint gl2Program = glCreateProgram();
  
  /* ֥ȤΥץؤϿ */
  glAttachShader(gl2Program, vertShader);
  glAttachShader(gl2Program, fragShader);
  
  /* ֥Ȥκ */
  glDeleteShader(vertShader);
  glDeleteShader(fragShader);
  
  /* ץΥ */
  glLinkProgram(gl2Program);
  glGetProgramiv(gl2Program, GL_LINK_STATUS, &status);
  printProgramInfoLog(gl2Program);
  if (status == GL_FALSE) {
    fprintf(stderr, "Link error.\n");
    return 0;
  }
  
  return gl2Program;
}

/*
** 
*/
static void cleanup(void)
{
  delete obj;
}

/*
** 
*/
static void init(void)
{
#if defined(WIN32)
  /* GLEW ν */
  GLenum err = glewInit();
  if (err != GLEW_OK) {
    fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
    exit(1);
  }
#endif
  /* ץɤ߹ */
  shader = loadShader(vertSource, fragSource);
  if (shader == 0) exit(1);
  
  /* ǡɤ߹ */
  obj = new Obj;
  if (obj == 0) exit(1);
  atexit(cleanup);
  if (!obj->load(objectData)) exit(1);

  /* طʿ */
  glClearColor(0.0, 0.1, 0.3, 1.0);
  
  /* в֤ν */
  glutGet(GLUT_ELAPSED_TIME);
}

/*
** ᥤץ
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();
  
  return 0;
}
